"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import sys
import json
import inspect
from abc import abstractmethod
import os
from enum import Enum
from typing import Any, Mapping, TypeVar, Callable, Optional
from pathlib import Path
from datetime import datetime, timezone
from importlib.metadata import PackageMetadata, metadata

# * Third Party Imports --------------------------------------------------------------------------------->
import psutil
from yarl import URL
from platformdirs import PlatformDirs

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.custom_types import PATH_TYPE
from gidapptools.utility.enums import NamedMetaPath
from gidapptools.general_helper.date_time import DatetimeFmt
from gidapptools.general_helper.conversion import str_to_bool

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.resolve()

# endregion [Constants]


class PackageMetadataDict(dict):
    list_fields: set[str] = {"dynamic",
                             "platform",
                             "supported-platform",
                             "classifier",
                             "requires-dist",
                             "requires-external",
                             "project-url",
                             "provides-extra",
                             "provides-dist",
                             "obsoletes-dist"}

    get_fallback_keys: dict[str, tuple[str]] = {"maintainer": ("author",)}

    @property
    def all_urls(self) -> dict[str, str]:
        _out = {}
        for name_url_pair in self.get("project-url", []):
            try:
                name, url = (i.strip() for i in name_url_pair.split(","))
                _out[name] = url
            except ValueError:
                continue

        for url_key in ["Download-URL", "Home-page"]:
            try:
                _out[url_key] = self[url_key.casefold()]
            except KeyError:
                continue

        return _out

    def add(self, k, v) -> None:
        k = k.casefold()

        if k in self.list_fields:
            try:
                v = self[k] + [v]
            except KeyError:
                v = [v]

        super().__setitem__(k, v)

    def __getitem__(self, k):
        try:
            return super().__getitem__(k)
        except KeyError as e:
            if k in self.get_fallback_keys:
                for fallback_k in self.get_fallback_keys[k]:
                    try:
                        return super().__getitem__(fallback_k)
                    except KeyError:
                        continue
        raise KeyError(k)

    @classmethod
    def from_meta_importlib_meta_data(cls, importlib_meta_data: PackageMetadata) -> "PackageMetadataDict":
        instance = cls()
        for k, v in importlib_meta_data.items():
            instance.add(k, v)
        return instance


def abstract_class_property(func):
    return property(classmethod(abstractmethod(func)))


def utc_now():
    return datetime.now(tz=timezone.utc)


def handle_path(path: Optional[PATH_TYPE]):
    if path is None:
        return path
    return Path(path).resolve()


def memory_in_use():
    memory = psutil.virtual_memory()
    return memory.total - memory.available


def meta_data_from_path(in_path: Path) -> dict[str, Any]:
    _init_module = inspect.getmodule(None, in_path)

    _metadata = metadata(_init_module.__package__)

    _out = PackageMetadataDict.from_meta_importlib_meta_data(_metadata)

    return _out


def meta_data_from_package_name(package_name: str) -> dict[str, Any]:

    _metadata = metadata(package_name)

    _out = PackageMetadataDict.from_meta_importlib_meta_data(_metadata)

    return _out


TCallable = TypeVar("TCallable", bound=Callable)


def _mark_appdir_path(func: TCallable) -> TCallable:
    func._appdir_path_type = NamedMetaPath(func.__name__)
    return func


class PathLibAppDirs:
    mark_path = _mark_appdir_path

    def __init__(self,
                 appname: str,
                 appauthor: str = None,
                 version: str = None,
                 roaming: bool = True,
                 multipath: bool = False) -> None:
        self.platform_dirs = PlatformDirs(appname=appname, appauthor=appauthor, version=version, roaming=roaming, multipath=multipath)

    @property
    def appname(self) -> str:
        return self.platform_dirs.appname

    @property
    def authorname(self) -> Optional[str]:
        return self.platform_dirs.appauthor

    @mark_path
    def user_data_dir(self) -> Path:
        return Path(self.platform_dirs.user_data_path)

    @mark_path
    def user_log_dir(self) -> Path:
        return Path(self.platform_dirs.user_log_path)

    @mark_path
    def user_cache_dir(self) -> Path:
        return Path(self.platform_dirs.user_cache_path)

    @mark_path
    def user_config_dir(self) -> Path:
        return Path(self.platform_dirs.user_config_path)

    @mark_path
    def user_state_dir(self) -> Path:
        return Path(self.platform_dirs.user_state_path)

    @mark_path
    def site_data_dir(self) -> Path:
        return Path(self.platform_dirs.user_data_path)

    @mark_path
    def site_config_dir(self) -> Path:

        return Path(self.platform_dirs.site_config_path)

    def as_path_dict(self) -> dict[NamedMetaPath, Optional[Path]]:
        path_dict = {named_path_item: None for named_path_item in NamedMetaPath.__members__.values()}
        for meth_name, meth_object in inspect.getmembers(self, inspect.ismethod):
            if hasattr(meth_object, '_appdir_path_type'):
                path_dict[meth_object._appdir_path_type] = meth_object()
        return path_dict


def make_pretty(inst) -> dict:

    # pylint: disable=too-many-return-statements
    def _make_pretty(obj):

        def _handle_iterable(in_obj):
            return [_make_pretty(item) for item in in_obj]

        if isinstance(obj, Enum):
            return str(obj)

        if isinstance(obj, Path):
            return obj.as_posix()
        if isinstance(obj, URL):
            return str(obj)
        if isinstance(obj, datetime):
            return DatetimeFmt.STANDARD.strf(obj)

        if isinstance(obj, Mapping):
            pretty_dict = {}
            for key, value in obj.items():
                pretty_dict[_make_pretty(key)] = getattr(inst, f"pretty_{key}", _make_pretty(value))
            return pretty_dict

        if isinstance(obj, list):
            return _handle_iterable(obj)

        if isinstance(obj, tuple):
            return tuple(_handle_iterable(obj))
        if isinstance(obj, set):
            return set(_handle_iterable(obj))
        if isinstance(obj, frozenset):
            return frozenset(_handle_iterable(obj))
        return obj
    return _make_pretty(vars(inst))


def get_qualname_or_name(in_object: Any) -> str:
    try:
        return in_object.__qualname__
    except AttributeError:
        return in_object.__name__


def merge_json_files(base_file: Path, file_to_merge: Path, content_type: type = dict):
    if content_type is not dict:
        raise TypeError(f"function not Implemented for type {content_type!r}")
    with base_file.open("r", encoding='utf-8', errors='ignore') as f_b:
        data_b = json.load(f_b)
    with file_to_merge.open("r", encoding='utf-8', errors='ignore') as f_m:
        data_m = json.load(f_m)

    combined = data_b | data_m

    with base_file.open("w", encoding='utf-8', errors='ignore') as f:
        json.dump(combined, f, default=str, sort_keys=False)


def merge_content_to_json_file(base_file: Path, content: str):
    base_data = json.loads(base_file.read_text(encoding='utf-8', errors='ignore'))
    merge_data = json.loads(content)
    combined_data = base_data | merge_data
    with base_file.open("w", encoding='utf-8', errors='ignore') as f:
        json.dump(combined_data, f, default=str, sort_keys=False, indent=4)


def get_main_module_path() -> Path:
    main_module = sys.modules["__main__"]

    return Path(main_module.__file__).resolve()


def is_dev() -> bool:
    is_dev_string = os.getenv("IS_DEV", '0').casefold()
    return str_to_bool(is_dev_string, strict=True) or sys.flags.dev_mode


# region [Main_Exec]
if __name__ == '__main__':
    pass
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import TYPE_CHECKING, Hashable, Optional
from pathlib import Path
from threading import Lock

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.gid_signal.signal_registry import signal_registry
from gidapptools.gid_signal.signals.basic_signal import GidSignal

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.gid_signal.signals.abstract_signal import AbstractSignal

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]

GET_SIGNAL_LOCK = Lock()


def get_signal(key: Hashable, klass: Optional["AbstractSignal"] = None) -> "AbstractSignal":
    klass = GidSignal if klass is None else klass
    with GET_SIGNAL_LOCK:
        signal = signal_registry.get(key, None)
        if signal is None:
            signal = klass(key=key)
            signal_registry.register(signal)

    return signal


# region [Main_Exec]

if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from pathlib import Path
from threading import Lock, RLock
from weakref import WeakValueDictionary
from filelock import FileLock
# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.custom_types import LOCK_CLASS_TYPE, PATH_TYPE, LOCK_TYPE

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class FileLocksManager:
    __slots__ = ("_lock_type",
                 "_interaction_lock",
                 "_file_locks")

    def __init__(self, lock_type: LOCK_CLASS_TYPE):
        self._lock_type = lock_type
        self._interaction_lock: Lock = Lock()
        self._file_locks: WeakValueDictionary[Path, LOCK_TYPE] = WeakValueDictionary()

    def _handle_file_path(self, file_path: PATH_TYPE) -> Path:
        return Path(file_path).resolve()

    def _get_or_create(self, file_path: PATH_TYPE) -> LOCK_CLASS_TYPE:
        file_path = self._handle_file_path(file_path=file_path)
        with self._interaction_lock:
            try:
                return self._file_locks[file_path]
            except KeyError:
                file_lock = self._lock_type()
                self._file_locks[file_path] = file_lock
                return file_lock

    def get_file_lock(self, file_path: PATH_TYPE) -> LOCK_CLASS_TYPE:
        return self._get_or_create(file_path=file_path)

    def __getitem__(self, file_path: PATH_TYPE) -> LOCK_CLASS_TYPE:
        return self.get_file_lock(file_path=file_path)

    def __len__(self) -> int:
        with self._interaction_lock:
            return len(self._file_locks)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(lock_type={self._lock_type!r})"


GLOBAL_LOCK_MANAGER = FileLocksManager(Lock)
GLOBAL_RLOCK_MANAGER = FileLocksManager(RLock)


# region [Main_Exec]
if __name__ == '__main__':
    pass
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from pathlib import Path

# * Third Party Imports --------------------------------------------------------------------------------->
import pyparsing as pp

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.gid_config.parser.tokens import TokenFactory

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class BaseIniGrammar:
    """
    Builds the pyparsing Grammar for the Parser.

    Args:
        key_value_separator (str, optional): The string that indicates the separation between a key and its value, gets also used for transforming back into text. Defaults to '='.
        comment_indicator (str, optional): String that indicates the start of a comment, gets also used for transforming back into text. Defaults to '#'.
        token_factory (TokenFactory, optional): Factory that transforms pyparsing results into tokens, gets set as `set_parse_action`s. Defaults to None.
    """
    __slots__ = ("token_factory", "raw_key_value_separator", "key_value_separator", "raw_comment_indicator", "comment_indicator")
    l_sqr_bracket = pp.Literal("[").suppress()
    r_sqr_bracket = pp.Literal("]").suppress()

    section_name_excluded_chars = ['[', ']', '.']
    section_name_extra_chars = ['\t']

    key_name_exclusion_chars = ['[', ']', '.']
    key_name_extra_chars = [" "]

    value_exclusion_chars = [" "]
    value_extra_chars = [' ', '\t']
    base_chars = {"section": {"exclude": section_name_excluded_chars,
                              "extra": section_name_extra_chars},
                  "key": {"exclude": key_name_exclusion_chars,
                          "extra": key_name_extra_chars},
                  "value": {"exclude": value_exclusion_chars,
                            "extra": value_extra_chars}}

    def __init__(self,
                 key_value_separator: str = '=',
                 comment_indicator: str = '#',
                 token_factory: TokenFactory = None) -> None:
        self.token_factory = TokenFactory() if token_factory is None else token_factory
        self.raw_key_value_separator = key_value_separator
        self.key_value_separator = pp.Suppress(self.raw_key_value_separator)
        self.raw_comment_indicator = comment_indicator
        self.comment_indicator = pp.Suppress(self.raw_comment_indicator)

    def get_chars_for(self, kind: str) -> str:
        exclude = self.base_chars[kind]['exclude'] + [self.raw_key_value_separator, self.raw_comment_indicator]
        extra = self.base_chars[kind]['extra']
        return ''.join(char for char in pp.printables if char not in exclude) + ''.join(extra)

    @property
    def section_name(self) -> pp.ParserElement:
        section_name = pp.AtLineStart(self.l_sqr_bracket + pp.Word(self.get_chars_for('section')) + self.r_sqr_bracket)
        return section_name.set_results_name('section_name')

    @property
    def key(self) -> pp.ParserElement:
        key = pp.AtLineStart(pp.Word(init_chars=self.get_chars_for("key").replace('[', ''), body_chars=self.get_chars_for('key'))) + self.key_value_separator
        return key

    @property
    def value(self) -> pp.ParserElement:
        value = pp.OneOrMore(pp.Word(init_chars=self.get_chars_for("value").replace('[', ''), body_chars=self.get_chars_for('value')),
                             stop_on=self.key | self.section_name | self.comment).set_parse_action(''.join)

        return value

    @ property
    def entry(self) -> pp.ParserElement:
        entry = self.key + pp.Optional(self.value)
        return entry.set_results_name('entry')

    @ property
    def comment(self) -> pp.ParserElement:
        comment = self.comment_indicator + pp.rest_of_line
        return comment.set_results_name('comment')

    def get_grammar(self, **kwargs) -> pp.ParserElement:
        all_elements = pp.MatchFirst([self.section_name, self.entry, self.comment])
        # all_elements = self.section_name | self.entry | self.comment
        return all_elements.set_parse_action(self.token_factory.parse_action)


# region [Main_Exec]

if __name__ == '__main__':
    pass
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from typing import Union, TypedDict
from pathlib import Path

# * Third Party Imports --------------------------------------------------------------------------------->
import pyparsing as ppa
from pyparsing import common as ppc

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.errors import InvalidConverterValue
from gidapptools.general_helper.timing import get_dummy_profile_decorator_in_globals

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]
get_dummy_profile_decorator_in_globals()
THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


REPLACE_WORDS = {"$comma$": ",",
                 "$equals$": "="}


def replace_value_words(in_tok: ppa.ParseResults) -> str:
    text = str(in_tok[0])
    for k, v in REPLACE_WORDS.items():
        text = text.replace(k, v)

    return text


def reverse_replace_value_words(in_value: str) -> str:
    text = str(in_value)
    for k, v in REPLACE_WORDS.items():
        text = text.replace(v, k)
    return text


def get_converter_specification_grammar() -> ppa.ParserElement:
    parenthesis_open = ppa.Literal("(").suppress()
    parenthesis_close = ppa.Literal(")").suppress()

    equal_sign = ppa.Literal("=").suppress()
    comma = ppa.Literal(",").suppress()

    word_value = ppa.Word(''.join(c for c in ppa.printables if c not in {"=", ",", "(", ")"})).set_parse_action(replace_value_words)
    bool_value = ppa.CaselessLiteral("true").set_parse_action(lambda x: True) | ppa.CaselessKeyword("false").set_parse_action(lambda x: False)
    kw_value = ppc.number | bool_value | word_value

    kw_argument = ppa.Group(ppc.identifier + equal_sign + kw_value)

    arguments = ppa.Dict(ppa.ZeroOrMore(kw_argument + ppa.Opt(comma)), asdict=True).set_parse_action(lambda x: x[0])("kw_arguments")

    typus = ppa.Word(ppa.alphas + "_", ppa.alphanums + '_')

    sub_arguments = parenthesis_open + arguments + parenthesis_close

    grammar = typus("typus") + ppa.Opt(sub_arguments)

    return grammar


CONVERTER_SPECIFICATION_GRAMMAR = get_converter_specification_grammar()


class ConverterSpecData(TypedDict):
    typus: str
    kw_arguments: dict[str, Union[str, int, float]]


def parse_specification(raw_specification: str) -> ConverterSpecData:
    try:
        result = CONVERTER_SPECIFICATION_GRAMMAR.parse_string(raw_specification, parse_all=True).as_dict()
        return ConverterSpecData(typus=result["typus"], kw_arguments=result.get("kw_arguments", dict()))
    except ppa.ParseException as error:
        raise InvalidConverterValue(f"{raw_specification!r} is not a valid converter specification") from error


# region [Main_Exec]
if __name__ == '__main__':
    pass
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import sys
from typing import TYPE_CHECKING
from pathlib import Path
from threading import RLock

if sys.version_info >= (3, 11):
    pass
else:
    pass
# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    ...

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class Switch:
    __slots__ = ("_lock", "_state")

    def __init__(self, initial_state: bool = False) -> None:
        self._lock = RLock()
        self._state = initial_state

    @property
    def state(self) -> bool:
        with self._lock:
            return self._state

    @property
    def is_true(self) -> bool:
        with self._lock:
            return self.state is True

    @property
    def is_false(self) -> bool:
        with self._lock:
            return self.state is False

    def switch_to(self, new_state: bool) -> None:
        with self._lock:
            self._state = new_state

    def switch(self) -> bool:
        with self._lock:
            self.switch_to(not self.state)
            return self.state

    def __bool__(self) -> bool:
        return self.state

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(value={self.state!r})"

# region [Main_Exec]


if __name__ == '__main__':
    pass


# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import asyncio
from time import sleep
from typing import Union, Callable
from pathlib import Path
from functools import partial
from concurrent.futures import ThreadPoolExecutor

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.utility.helper import get_qualname_or_name

from .abstract_signal import AbstractSignal

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


def run_delayed(delay: Union[int, float], emit_func: Callable):
    sleep(delay)
    emit_func()


class GidSignal(AbstractSignal):

    def fire_and_forget(self, *args, **kwargs):
        if len(self.targets) <= 0:
            return
        with ThreadPoolExecutor(thread_name_prefix='signal_thread') as pool:
            for target in self.targets:

                pool.submit(target, *args, **kwargs)

    def delayed_fire_and_forget(self, delay: Union[int, float], *args, **kwargs):

        if len(self.targets) <= 0:
            return
        with ThreadPoolExecutor(thread_name_prefix='signal_thread') as pool:
            for target in self.targets:
                emit_func = partial(target, *args, **kwargs)
                pool.submit(run_delayed, delay=delay, emit_func=emit_func)

    def emit(self, *args, **kwargs):

        for target in self.targets:
            target(*args, **kwargs)

    async def aemit(self, *args, **kwargs):
        if len(self.targets) <= 0:
            return

        for target in self.targets:
            name = get_qualname_or_name(target)
            info = self.targets_info.get(name)
            task_name = f"{str(self.key)}-Signal_{name}"
            if info.get('is_coroutine') is False:
                task = asyncio.to_thread(target, *args, **kwargs)
            else:
                task = target(*args, **kwargs)
            asyncio.create_task(task, name=task_name)


# region [Main_Exec]
if __name__ == '__main__':
    pass
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import os
from typing import TYPE_CHECKING, Any, Union, Literal, Iterable
from pathlib import Path
from threading import RLock

# * Gid Imports ----------------------------------------------------------------------------------------->
from gidapptools.errors import EntryMissingError, MissingDefaultValue
from gidapptools.general_helper.enums import MiscEnum
from gidapptools.gid_signal.interface import get_signal
from gidapptools.general_helper.timing import get_dummy_profile_decorator_in_globals
from gidapptools.gid_config.parser.tokens import Section
from gidapptools.gid_config.parser.ini_parser import BaseIniParser
from gidapptools.gid_config.parser.config_data import ConfigFile
from gidapptools.gid_config.conversion.spec_data import SpecFile, SpecEntry, SpecLoader, SpecSection
from gidapptools.gid_config.conversion.conversion_table import ConversionTable, ConfigValueConverter
from gidapptools.gid_config.conversion.converter_grammar import ConverterSpecData

# * Type-Checking Imports --------------------------------------------------------------------------------->
if TYPE_CHECKING:
    from gidapptools.custom_types import PATH_TYPE

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]
get_dummy_profile_decorator_in_globals()
THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


class SectionAccessor:

    def __init__(self, config: "GidIniConfig", section_name: str) -> None:
        self.config = config
        self.section_name = section_name

    def get(self,
            entry_key: str,
            fallback_entry: Iterable[str] = None,
            default: Any = MiscEnum.NOTHING) -> Any:
        return self.config.get(section_name=self.section_name, entry_key=entry_key, fallback_entry=fallback_entry, default=default)

    def set(self,
            entry_key: str,
            entry_value: Any,
            create_missing_section: bool = False,
            spec_typus: str = None) -> None:
        return self.config.set(section_name=self.section_name, entry_key=entry_key, entry_value=entry_value, create_missing_section=create_missing_section, spec_typus=spec_typus)


class ResolvedSection:
    __slots__ = ("config", "name", "item", "spec")

    def __init__(self, config: "GidIniConfig", section_name: str, section_item: Section, spec_Section: SpecSection) -> None:
        self.config = config
        self.name = section_name
        self.item = section_item
        self.spec = spec_Section

    @property
    def entries(self) -> tuple["ResolvedEntry"]:
        _out = []
        for entry in self.spec.entries.values():

            _out.append(self.config.get_entry_item(self.name, entry.name))

        return tuple(_out)

    def __getattr__(self, name: str):
        for sub_item in [self.spec, self.item]:
            try:
                return getattr(sub_item, name)
            except AttributeError:
                pass
        raise AttributeError(name)

    def __repr__(self) -> str:

        return f'{self.__class__.__name__}(section_name={self.name!r})'


class ResolvedEntry:
    __slots__ = ("section_name", "entry_name", "spec_entry", "converter", "raw_value")

    def __init__(self,
                 section_name: str,
                 entry_name: str,
                 spec_entry: SpecEntry,
                 converter: ConfigValueConverter,
                 raw_value: str = MiscEnum.NOTHING) -> None:
        self.section_name = section_name
        self.entry_name = entry_name
        self.spec_entry = spec_entry
        self.converter = converter
        self.raw_value = raw_value

    @property
    def value_typus(self) -> str:
        return self.converter.value_typus

    @property
    def description(self) -> str:
        return self.spec_entry.description

    @property
    def verbose_name(self) -> str:
        return self.spec_entry.verbose_name

    @property
    def gui_visible(self) -> bool:
        return self.spec_entry.gui_visible

    @property
    def implemented(self) -> bool:
        return self.spec_entry.implemented

    @property
    def default(self) -> str:
        return self.spec_entry.default

    @property
    def initial_value(self) -> Any:
        if self.spec_entry.initial_value is MiscEnum.NOTHING:
            return MiscEnum.NOTHING

        return self.converter.to_python_value(self.spec_entry.initial_value)

    @property
    def value(self) -> Any:
        _value = self.raw_value

        if _value in {MiscEnum.NOTHING, None} and self.default is not MiscEnum.NOTHING:
            _value = self.spec_entry.default
        if _value is MiscEnum.NOTHING:
            raise MissingDefaultValue(f"No value or default value found for '{self.section_name}.{self.entry_name}'.")

        return self.converter.to_python_value(_value)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(section_name={self.section_name!r}, entry_name={self.entry_name!r})'


class GidIniConfig:
    access_locks_storage: dict[tuple, RLock] = {}

    __slots__ = ("spec",
                 "config",
                 "conversion_table",
                 "empty_is_missing",
                 "_resolve_entry_cache",
                 "changed_signal",
                 "__weakref__")

    def __init__(self,
                 spec_file: SpecFile,
                 config_file: ConfigFile,
                 conversion_table: ConversionTable,
                 empty_is_missing: bool = True) -> None:
        self._resolve_entry_cache: dict[tuple[str, str], ResolvedEntry] = {}
        self.spec = spec_file
        self.config = config_file
        self.conversion_table = conversion_table
        self.empty_is_missing = empty_is_missing
        self.changed_signal = get_signal(key=self.config.name)
        self.spec.changed_signal.connect(self.on_any_changed)
        self.config.add_on_reload_target(self.on_any_changed)

    @property
    def name(self) -> str:
        return self.config.name

    @property
    def is_dev(self) -> bool:
        return os.getenv("is_dev", "false") != "false"

    @property
    def sections(self) -> tuple[ResolvedSection]:
        return tuple(self.get_section(i) for i in self.spec.sections)

    @property
    def section_names(self) -> tuple[str]:
        return tuple(self.spec.sections)

    def clear_cache(self) -> None:
        self._resolve_entry_cache = {}

    def get_section(self, section_name: str) -> ResolvedSection:
        spec_section = self.spec.sections[section_name]
        section_item = self.config.get_section(section_name)

        return ResolvedSection(self, section_name, section_item, spec_section)

    def get_spec_item(self, section_name: str, entry_name: str) -> SpecEntry:
        return self.spec.get_spec_entry(section_name=section_name, entry_name=entry_name)

    def get_converter(self, converter_data: ConverterSpecData) -> ConfigValueConverter:
        return self.conversion_table.get_converter(converter_data=converter_data)

    def get_entry_item(self, section_name: str, entry_name: str) -> ResolvedEntry:

        try:
            return self._resolve_entry_cache[(section_name, entry_name)]
        except KeyError:
            pass

        spec_item = self.get_spec_item(section_name=section_name, entry_name=entry_name)
        converter = self.get_converter(spec_item.converter)

        entry_item = ResolvedEntry(section_name=section_name, entry_name=entry_name, spec_entry=spec_item, converter=converter)
        try:
            entry = self.config.get_entry(section_name=section_name, entry_key=entry_name)
            entry_item.raw_value = entry.value
        except EntryMissingError:
            pass
        self._resolve_entry_cache[(section_name, entry_name)] = entry_item
        return entry_item

    def get(self, section_name: str, entry_name: str, default=MiscEnum.NOTHING) -> Any:
        try:
            entry_item = self.get_entry_item(section_name=section_name, entry_name=entry_name)
            return entry_item.value
        except MissingDefaultValue:
            if default is not MiscEnum.NOTHING:
                return default
            raise

    def set(self, section_name: str, entry_name: str, value: Any) -> None:
        spec_item = self.get_spec_item(section_name=section_name, entry_name=entry_name)
        converter = self.get_converter(spec_item.converter)
        self.config.set_value(section_name=section_name, entry_key=entry_name, entry_value=converter.to_config_value(value))
        self.clear_cache()
        self.reload_if_changed()

    def ensure_section(self, section_name: str) -> None:
        section = Section(section_name)
        self.config.add_section(section)

    def set_if_not_exists(self, section_name: str, entry_name: str, value: Any) -> None:
        existing = self.config.has_key(section_name=section_name, key_name=entry_name)
        if existing is False:
            self.set(section_name=section_name, entry_name=entry_name, value=value)

    def reload_if_changed(self) -> None:
        self.spec.reload_if_changed()
        self.config.reload_if_changed()

    def reload(self) -> None:
        self.config.reload()
        self.spec.reload()

    def on_any_changed(self, changeling: Union[ConfigFile, SpecFile] = None) -> None:
        self.clear_cache()
        self.changed_signal.emit(self)

    def __repr__(self) -> str:
        """
        Basic Repr
        !REPLACE!
        """
        return f'{self.__class__.__name__}(name={self.name!r}, spec={self.spec!r}, config={self.config!r}, converter={self.conversion_table!r}, empty_is_missing={self.empty_is_missing!r})'


def preload_config(in_config: "GidIniConfig") -> None:
    for section in in_config.sections:
        in_config.ensure_section(section.name)
        if section.spec.dynamic_entries_allowed is False:
            for entry in section.entries:
                try:
                    if in_config.config.file_was_created is True and entry.initial_value is not MiscEnum.NOTHING:
                        value = entry.initial_value
                    else:
                        value = entry.value
                    in_config.set_if_not_exists(section.name, entry.entry_name, value)
                except MissingDefaultValue:
                    continue
    in_config.config.save()


def get_config(spec_path: "PATH_TYPE",
               config_path: "PATH_TYPE",
               spec_loader: SpecLoader = None,
               config_parser: BaseIniParser = None,
               config_auto_write: bool = True,
               changed_parameter: Union[Literal['size'], Literal['file_hash'], Literal["mtime"], Literal["never"], Literal["always"], Literal["all"]] = 'mtime',
               extra_converter: Iterable[ConfigValueConverter] = None,
               empty_is_missing: bool = True,
               preload_ini_file: bool = False) -> GidIniConfig:

    conversion_table = ConversionTable(extra_converter=extra_converter)
    spec = SpecFile(spec_path, loader=spec_loader or SpecLoader(), changed_parameter=changed_parameter)
    config = ConfigFile(config_path, parser=config_parser or BaseIniParser(), changed_parameter=changed_parameter, auto_write=config_auto_write)
    config_item = GidIniConfig(spec, config, conversion_table=conversion_table, empty_is_missing=empty_is_missing)

    if preload_ini_file is True:
        preload_config(config_item)
    return config_item


# region [Main_Exec]
if __name__ == '__main__':
    pass
# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
import json
import shutil
import inspect
from typing import Any, Union, Callable, Iterable, Optional
from pathlib import Path
from datetime import datetime
from collections import defaultdict
from gidapptools.gid_logger.logger import get_logger

# * Qt Imports --------------------------------------------------------------------------------------->
from PySide6.QtGui import QMouseEvent, QPaintEvent, QResizeEvent, QStatusTipEvent, QPlatformSurfaceEvent, QInputMethodQueryEvent
from PySide6.QtCore import QEvent, QChildEvent, QDynamicPropertyChangeEvent

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()
log = get_logger(__name__)
# endregion [Constants]


class JsonOutputter:

    def __init__(self, target_folder: Path, file_stem_suffix: str = None) -> None:
        self.target_folder = target_folder
        self.specific_folder = self.target_folder.joinpath("event_specific_data")
        self.target_folder.mkdir(exist_ok=True, parents=True)
        if self.specific_folder.exists() is True:
            shutil.rmtree(self.specific_folder)
        self.specific_folder.mkdir(exist_ok=True, parents=True)
        self.file_stem_suffix = file_stem_suffix

    def __call__(self, data: dict):
        def _clean_event_name(_event: QEvent) -> str:
            parts = str(_event.type()).split('.')
            while parts[0].casefold() != "qevent":
                _ = parts.pop(0)
            name = '_'.join(parts)
            return name

        event = data.get("event")
        specific = {"event": data.get("event"), "specific": data.pop("specific")}
        file_name = _clean_event_name(event)
        if self.file_stem_suffix is not None:
            file_name += f"_{self.file_stem_suffix}"
        file_name += '.json'
        path = self.target_folder.joinpath(file_name)
        if path.is_file() is False:

            with path.open("w", encoding='utf-8', errors='ignore') as f:
                json.dump(data, f, default=str, sort_keys=False, indent=4)

        if specific["specific"]:
            specific_file_name = f"{path.stem}_{datetime.now().strftime('%Y-%m-%d_%H-%M-%S-%f')}.json"
            specific_path = self.specific_folder.joinpath(specific_file_name)
            with specific_path.open("w", encoding='utf-8', errors='ignore') as f:
                json.dump(specific, f, default=str, sort_keys=False, indent=4)


def _data_getter_class_name(obj: object) -> str:
    try:
        if inspect.isclass(obj):
            return obj.__name__
    except Exception as e:
        log.error(e, exc_info=True)

    try:
        return obj.__class__.__name__
    except Exception as e:
        log.error(e, exc_info=True)

    return "Not able to determine class Name".upper()


def _data_getter_all_subclasses(obj: object) -> tuple[type]:

    if inspect.isclass(obj):
        klass = obj
    else:
        klass = obj.__class__
    all_subclasses = []

    for subclass in klass.__subclasses__():
        all_subclasses.append(subclass)
        all_subclasses.extend(_data_getter_all_subclasses(subclass))

    return tuple(all_subclasses)


def _data_getter_all_subclasses_names(obj: object) -> tuple[str]:
    return tuple(i.__name__ for i in _data_getter_all_subclasses(obj))


def _data_getter_all_member_names(obj: object) -> tuple[str]:
    return tuple(name for name, obj in inspect.getmembers(obj))


def _data_getter_all_non_param_methods(obj: object) -> tuple[str]:
    meths = []
    for name, obj in inspect.getmembers(obj):
        try:
            if len(inspect.getfullargspec(obj)[0]) <= 1:
                meths.append(name)
        except TypeError:
            continue
    return tuple(meths)


DEFAULT_SPEC_DATA_TO_GET = {QChildEvent: ["polished"],
                            QDynamicPropertyChangeEvent: ["propertyName"],
                            QResizeEvent: ["size", "oldSize"],
                            QStatusTipEvent: ["tip"],
                            QInputMethodQueryEvent: ["queries"],
                            QPlatformSurfaceEvent: ["surfaceEventTypes"],
                            QPaintEvent: ["region", "rect"],
                            QMouseEvent: ["source", "flags"]}

DEFAULT_SPEC_DATA_TO_GET = {k.__name__: v for k, v in DEFAULT_SPEC_DATA_TO_GET.items()}


class EventAnalyzer:
    std_data_to_get: tuple[str] = ("type", "_data_getter_class_name", "_data_getter_all_non_param_methods", "_data_getter_all_subclasses_names", "_data_getter_all_member_names", "spontaneous", "isSinglePointEvent", "isPointerEvent", "isInputEvent")
    specific_data_to_get: dict[str, list] = defaultdict(list, DEFAULT_SPEC_DATA_TO_GET)
    data_getters: dict[str, Callable] = {f.__name__: f for f in [_data_getter_all_member_names, _data_getter_all_subclasses, _data_getter_class_name, _data_getter_all_subclasses_names, _data_getter_all_non_param_methods]}

    def __init__(self, output: Callable = print, output_data: bool = False, only: Iterable[str] = None, exclude: Iterable[str] = None):
        self.output = output
        self.output_data = output_data
        self.only = set(only) if only is not None else only
        self.exclude = set(exclude) if exclude is not None else set()

    def _get_event_data(self, event: QEvent) -> dict[str, Any]:
        std_data = {}
        specific_data = {}
        for to_get in self.std_data_to_get:
            if to_get.startswith("_data_getter_"):
                std_data[to_get.removeprefix("_data_getter_")] = self.data_getters[to_get](event)
            else:
                value = getattr(event, to_get)
                if callable(value):
                    value = value()
                std_data[to_get] = value
        std_data = {k: v for k, v in sorted(std_data.items(), key=lambda x: len(x[0]))}
        for spec_to_get in self.specific_data_to_get[event.__class__.__name__]:
            value = getattr(event, spec_to_get)
            if callable(value):
                value = value()
            specific_data[spec_to_get] = value
        specific_data = {k: v for k, v in sorted(specific_data.items(), key=lambda x: len(x[0]))}
        return {"event": event, "std": std_data, "specific": specific_data}

    def _render_to_string(self, data: dict[str, Any]):
        return pp.fmt(data)

    def analyze(self, event: QEvent) -> Optional[Union[str, dict]]:
        if self.only is not None and event.__class__.__name__ not in self.only:
            return
        if event.__class__.__name__ in self.exclude:

            return
        data = self._get_event_data(event)
        if self.output_data is True:
            return data
        return self._render_to_string(data)

    def __call__(self, event: QEvent) -> None:
        text_or_data = self.analyze(event)
        if text_or_data is not None:
            self.output(text_or_data)


# region [Main_Exec]
if __name__ == '__main__':
    pass

# endregion [Main_Exec]



"""
WiP.

Soon.
"""

# region [Imports]

# * Standard Library Imports ---------------------------------------------------------------------------->
from pathlib import Path
from typing import TypeVar
from collections.abc import Mapping, Iterable
# * Qt Imports --------------------------------------------------------------------------------------->
from PySide6.QtGui import QScreen
from PySide6.QtWidgets import QWidget, QApplication

# endregion [Imports]

# region [TODO]


# endregion [TODO]

# region [Logging]


# endregion [Logging]

# region [Constants]

THIS_FILE_DIR = Path(__file__).parent.absolute()

# endregion [Constants]


def center_window(window: QWidget, allow_window_resize: bool = True) -> QWidget:
    app = QApplication.instance()
    if allow_window_resize is True:
        window.resize(window.sizeHint())

    screen = app.primaryScreen()
    screen_geo = QScreen.availableGeometry(screen)
    screen_center = screen_geo.center()

    window_geo = window.frameGeometry()
    window_geo.moveCenter(screen_center)
    window.move(window_geo.topLeft())
    return window


T_WIDGET = TypeVar("T_WIDGET", bound=QWidget)


def batch_modify_widget(widget: T_WIDGET, **kwargs) -> T_WIDGET:
    for method_name, method_arguments in kwargs.items():
        method = getattr(widget, method_name)

        if isinstance(method_arguments, Mapping):
            method(**method_arguments)

        elif isinstance(method_arguments, Iterable) and not isinstance(method_arguments, str):
            method(*method_arguments)

        elif method_arguments is ...:
            method()

        else:
            method(method_arguments)

    return widget

# region [Main_Exec]


if __name__ == '__main__':
    pass

# endregion [Main_Exec]
